#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Any, List

from mesh.kinds import Document, Paging, Page
from mesh.macro import mpi, spi


@spi("mesh")
class DataHouse(ABC):

    @abstractmethod
    @mpi("mesh.dh.writes")
    async def writes(self, docs: List[Document]):
        """
        batch write log message
        :param docs:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.dh.write")
    async def write(self, doc: Document):
        """
        write log message
        :param doc:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.dh.read")
    async def read(self, index: Paging) -> Page[Any]:
        """
        write log message
        :param index:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.dh.indies")
    async def indies(self, index: Paging) -> Page[Any]:
        """
        Export index list.
        :param index:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.dh.tables")
    async def tables(self, index: Paging) -> Page[Any]:
        """
        Export table list.
        :param index:
        :return:
        """
        pass



#
# Copyright (c) 2019, 2023, ducesoft and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import asyncio
import os

import mesh


async def main():
    os.environ.setdefault("mesh.mode", str(1 << 6))
    os.environ.setdefault("mesh.proc", str(3))
    os.environ.setdefault("mesh.address", "10.43.245.145:7304")
    os.environ.setdefault("mesh.runtime", "192.168.31.11:9999")
    await mesh.start()
    await mesh.wait()


if __name__ == '__main__':

    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Dict, List

from mesh.macro import spi


@spi("mesh")
class Hodor(ABC):

    @abstractmethod
    async def stats(self, features: List[str]) -> Dict[str, str]:
        """
        Collect the system, application, process or thread status etc.
        :param features: customized parameters
        :return: quota stat
        """
        pass

    @abstractmethod
    async def debug(self, features: Dict[str, str]):
        """
        Set debug features.
        :param features:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any

from mesh.environ import Mode
from mesh.macro import spi
from mesh.mpc.filter import Filter, Invoker, Invocation, CONSUMER


@spi(name="isolate", pattern=CONSUMER, priority=100 - (1 << 32))
class IsolateFilter(Filter):

    async def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        if Mode.Isolate.enable() and self.is_isolate(invocation):
            return None
        return await invoker.run(invocation)

    @staticmethod
    def is_isolate(invocation: Invocation) -> bool:
        return "mesh.net.environ,mesh.registry.put,mesh.registry.puts,mesh.registry.remove".__contains__(
            invocation.get_urn().name)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, Dict

import mesh.log as log
from mesh.cause import MeshCode, Codeable, MeshException
from mesh.codec import Codec, Json
from mesh.context import Mesh
from mesh.kinds import Location
from mesh.macro import spi, ServiceLoader
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, PROVIDER
from mesh.prsim import Metadata


@spi(name="provider", pattern=PROVIDER, priority=(1 << 32) - 1)
class ProviderFilter(Filter):

    async def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        codec = ServiceLoader.load(Codec).get(Json)

        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not attachments:
            attachments = {}
        attachments[Metadata.MESH_CONSUMER.key()] = attachments[Metadata.MESH_PROVIDER.key()]
        attachments[Metadata.MESH_PROVIDER.key()] = codec.encode_string(await Location.localize(Location()))

        Mesh.context().decode(attachments)

        digest = Digest()
        try:
            ret = await invoker.run(invocation)
            digest.write("P", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("P", e.get_code())
            else:
                digest.write("P", MeshCode.SYSTEM_ERROR.get_code())
            if isinstance(e, MeshException):
                log.error(f"{digest.trace_id},{Mesh.context().get_urn()},{e.get_message}")
            raise e



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod

from mesh.kinds import Paging, Page
from mesh.macro import spi, mpi


@spi("mesh")
class IOStream(ABC):

    @abstractmethod
    @mpi(name="${mesh.name}.io.page.read", timeout=60000)
    async def read(self, index: Paging) -> Page[bytes]:
        """
        Read storage with page.
        :param index:
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="${mesh.name}.io.page.write", timeout=60000)
    async def write(self, index: Page[bytes]):
        """
        Write storage with page.
        :param index:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod, ABC
from typing import Any

from mesh.kinds import Reference
from mesh.macro import spi
from mesh.mpc.invoker import Execution
from mesh.mpc.urn import URN


@spi(name="grpc")
class Consumer(ABC):
    HTTP = "http"
    GRPC = "grpc"
    TCP = "tcp"
    MQTT = "mqtt"

    """
    Service consumer with any protocol and codec.
    """

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @abstractmethod
    def start(self):
        """
        Start the mesh broker.
        :return:
        """
        pass

    @abstractmethod
    async def consume(self, address: str, urn: URN, execution: Execution[Reference], inbound: bytes,
                      args: Any) -> bytes:
        """
        Consume the input payload.
        :param address: Remote address.
        :param urn: Actual uniform resource domain name.
        :param execution: Service reference.
        :param inbound: Input arguments.
        :param args: Additional arguments.
        :return: Output payload
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Dict, Any, Callable, Coroutine

import mesh.log as log
import mesh.tool as tool
from mesh.kinds import Principal
from mesh.macro import spi
from mesh.mpc import ServiceProxy, Mesh
from mesh.prsim import Transport, Session, Metadata


@spi("mesh")
class MeshTransport(Transport):
    """
    Mesh transport is stateless transport depends on mesh server.
    """

    def __init__(self):
        self.sessions: Dict[str, Session] = dict()

    async def open(self, session_id: str, metadata: Dict[str, str]) -> Session:
        skey = self.session_key(session_id, metadata)
        session = self.sessions.get(skey, None)
        if not session:
            attachments = Mesh.context().get_attachments()
            addr = Mesh.context().get_attribute(Mesh.REMOTE)
            principal = Mesh.context().get_principals().peek()
            session = MeshSession(session_id, metadata, attachments, addr, principal, lambda: self.finalize(skey))
            if self.sessions.__len__() < 500:
                self.sessions[skey] = session
        return session

    async def close(self, timeout: int):
        for k, v in self.sessions.items():
            try:
                await v.release(timeout, "")
            except BaseException as e:
                log.error("", e)
        self.sessions.clear()

    async def roundtrip(self, payload: bytes, metadata: Dict[str, str]) -> bytes:
        """ bid """
        pass

    def finalize(self, session_key: str):
        self.sessions.__delitem__(session_key)

    @staticmethod
    def session_key(session_id: str, metadata: Dict[str, str]) -> str:
        target_node_id = Metadata.MESH_TARGET_NODE_ID.get(metadata)
        target_inst_id = Metadata.MESH_TARGET_INST_ID.get(metadata)
        return f"{session_id}.{target_node_id}.{target_inst_id}"


@spi("mesh")
class MeshSession(Session):

    def __init__(self, session_id: str, metadata: Dict[str, str], attachments: Dict[str, str], address: str,
                 principal: Principal, finalizer: Callable):
        self.session_id = session_id
        self.metadata = metadata
        self.attachments = attachments
        self.address = address
        self.principal = principal
        self.session = ServiceProxy.default_proxy(Session)
        self.finalizer = finalizer

    async def with_context(self, fn: Coroutine, remote: bool, timeout: int) -> Any:
        return await Mesh.context_safe(self.context_safe(fn, remote, timeout))

    async def context_safe(self, fn: Any, remote: bool, timeout: int) -> Any:
        if self.attachments:
            for k, v in self.attachments.items():
                Mesh.context().get_attachments()[k] = v

        if self.metadata:
            for k, v in self.metadata.items():
                Mesh.context().get_attachments()[k] = v

        if tool.required(self.address):
            Mesh.context().set_attribute(Mesh.REMOTE, self.address)

        if timeout > 0:
            Mesh.context().set_attribute(Mesh.TIMEOUT, timeout)

        Metadata.MESH_SESSION_ID.set(Mesh.context().get_attachments(), self.session_id)

        if not remote:
            return await fn

        target_node_id = Metadata.MESH_TARGET_NODE_ID.get(Mesh.context().get_attachments())
        target_inst_id = Metadata.MESH_TARGET_INST_ID.get(Mesh.context().get_attachments())
        if '' == target_node_id and '' == target_inst_id:
            return await fn
        try:
            principal = Principal()
            principal.node_id = target_node_id
            principal.inst_id = target_inst_id
            Mesh.context().get_principals().append(principal)
            return await fn
        finally:
            Mesh.context().get_principals().pop()

    async def peek(self, topic: str = "") -> bytes:
        return await self.with_context(self.session.peek(topic), False, 0)

    async def pop(self, timeout: int, topic: str = "") -> bytes:
        return await self.with_context(self.session.pop(timeout, topic), False, timeout)

    async def push(self, payload: bytes, metadata: Dict[str, str], topic: str = ""):
        return await self.with_context(self.session.push(payload, metadata, topic), True, 0)

    async def release(self, timeout: int, topic: str = ""):
        self.finalizer()
        return await self.with_context(self.session.release(timeout, topic), False, timeout)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.asm import *
from mesh.boost import *
from mesh.cause import *
from mesh.codec import *
from mesh.context import *
from mesh.environ import *
from mesh.log import *
from mesh.macro import *
from mesh.mpc import *
from mesh.prsim import *

__all__ = (
    "mpi",
    "mps",
    "index",
    "spi",
    "binding",
    "Codeable",
    "Cause",
    "Inspector",
    "Types",
    "ServiceLoader",
    ##
    "URN",
    "URNFlag",
    "Consumer",
    "Filter",
    "Invocation",
    "Provider",
    "ServiceProxy",
    "MeshKey",
    "Mesh",
    # cause
    "MeshException",
    "CompatibleException",
    "NotFoundException",
    "ValidationException",
    "Codec",
    # prsim
    "Builtin",
    "Cache",
    "Cluster",
    "Commercialize",
    "RunMode",
    "Key",
    "Metadata",
    "Queue",
    "Context",
    "Cryptor",
    "DataHouse",
    "Dispatcher",
    "Endpoint",
    "EndpointSticker",
    "Evaluator",
    "Graphics",
    "Hodor",
    "IOStream",
    "KV",
    "Licenser",
    "Locker",
    "Network",
    "Publisher",
    "Registry",
    "Routable",
    "RuntimeHook",
    "Scheduler",
    "Sequence",
    "Subscriber",
    "Tokenizer",
    "Transport",
    "Session",
    #
    "Runtime",
    "MethodProxy",
    "Metadata",
)

__mooter__ = Mooter()


def init():
    asm.init()


async def start():
    await __mooter__.start()


async def refresh():
    await __mooter__.refresh()


async def stop():
    await __mooter__.stop()


async def wait():
    """
    Use signal handler to throw exception which can be caught to allow graceful exit.
    """
    await __mooter__.wait()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import time

import mesh.log as log
import mesh.tool as tool
from mesh.context import Mesh


class Digest:

    def __init__(self):
        self.trace_id = Mesh.context().get_trace_id()
        self.span_id = Mesh.context().get_span_id()
        self.mode = Mesh.context().get_run_mode()
        self.cni = tool.anyone(Mesh.context().get_consumer().node_id, '')
        self.cii = tool.anyone(Mesh.context().get_consumer().inst_id, '')
        self.cip = tool.anyone(Mesh.context().get_consumer().ip, '')
        self.chost = tool.anyone(Mesh.context().get_consumer().host, '')
        self.pni = tool.anyone(Mesh.context().get_provider().node_id, '')
        self.pii = tool.anyone(Mesh.context().get_provider().inst_id, '')
        self.pip = tool.anyone(Mesh.context().get_provider().ip, '')
        self.phost = tool.anyone(Mesh.context().get_provider().host, '')
        self.urn = Mesh.context().get_urn()
        self.now = int(time.time() * 1000)

    def write(self, pattern: str, code: str):
        now = int(time.time() * 1000)
        log.info(
            f"{self.trace_id},"
            f"{self.span_id},"
            f"{Mesh.context().get_timestamp()},"
            f"{self.now},"
            f"{now - Mesh.context().get_timestamp()},"
            f"{now - self.now},"
            f"{self.mode},"
            f"{pattern},"
            f"{self.cni},"
            f"{self.cii},"
            f"{self.pni},"
            f"{self.pii},"
            f"{self.cip},"
            f"{self.pip},"
            f"{self.chost},"
            f"{self.phost},"
            f"{tool.anyone(Mesh.context().get_attribute(Mesh.REMOTE), tool.get_mesh_address().any())},"
            f"{self.urn},{code}")



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import time
from typing import Dict, Any, Optional, Type

import mesh.log as log
import mesh.tool as tool
from mesh.codec import Codec, Json
from mesh.kinds import Principal, Location
from mesh.macro import ServiceLoader
from mesh.macro import T
from mesh.prsim import Key, RunMode, Queue, Context, Metadata


class MeshContext(Context):

    @staticmethod
    def create() -> Context:
        return MeshContext()

    def __init__(self):
        self.trace_id = tool.new_trace_id()
        self.span_id = tool.new_span_id("", 0)
        self.timestamp = int(time.time() * 1000)
        self.run_mode = RunMode.ROUTINE.value
        self.urn = ''
        self.calls = 0
        self.consumer = Location()
        self.provider = Location()
        self.attachments = dict()
        self.attributes = dict()
        self.principals = Queue()

    def get_trace_id(self) -> str:
        return self.trace_id

    def get_span_id(self) -> str:
        return self.span_id

    def get_timestamp(self) -> int:
        return self.timestamp

    def get_run_mode(self) -> int:
        return self.run_mode

    def get_urn(self) -> str:
        return self.urn

    def get_consumer(self) -> Location:
        return self.consumer

    def get_provider(self) -> Location:
        return self.provider

    def get_attachments(self) -> Dict[str, str]:
        return self.attachments

    def get_principals(self) -> Queue[Principal]:
        return self.principals

    def get_attributes(self) -> Dict[str, Any]:
        return self.attributes

    def get_attribute(self, key: Key[T]) -> T:
        return self.attributes.get(key.name, None)

    def set_attribute(self, key: Key[T], value: T) -> None:
        self.attributes[key.name] = value

    def rewrite_urn(self, urn: str) -> None:
        self.urn = urn

    def rewrite_context(self, another: Context) -> None:
        if tool.required(another.get_trace_id()):
            self.trace_id = another.get_trace_id()
        if tool.required(another.get_span_id()):
            self.span_id = another.get_span_id()
        if another.get_timestamp() > 0:
            self.timestamp = another.get_timestamp()
        if RunMode.ROUTINE.value != another.get_run_mode():
            self.run_mode = another.get_run_mode()
        if tool.required(another.get_urn()):
            self.urn = another.get_urn()
        if tool.required(another.get_consumer()):
            self.consumer = another.get_consumer()
        if tool.required(another.get_provider()):
            self.provider = another.get_provider()
        if tool.required(another.get_attachments()):
            for key, value in another.get_attachments().items():
                self.attachments[key] = value
        if tool.required(another.get_attributes()):
            for key, value in another.get_attributes().items():
                self.attributes[key] = value
        if tool.required(another.get_principals()):
            for value in another.get_principals():
                self.principals.append(value)

    def resume(self) -> Context:
        self.calls += 1
        context = MeshContext()
        context.rewrite_context(self)
        return context

    def encode(self) -> Dict[str, str]:
        return self.get_attachments()

    def decode(self, attachments: Dict[str, str]) -> None:
        if not attachments:
            attachments = {}
        codec = ServiceLoader.load(Codec).get(Json)
        trace_id = attachments.get(Metadata.MESH_TRACE_ID.key(), tool.new_trace_id())
        span_id = attachments.get(Metadata.MESH_SPAN_ID.key(), tool.new_span_id("", 0))
        timestamp = attachments.get(Metadata.MESH_TIMESTAMP.key(), '')
        run_mode = attachments.get(Metadata.MESH_RUN_MODE.key(), '')
        urn = attachments.get(Metadata.MESH_URN.key())
        cstr = attachments.get(Metadata.MESH_CONSUMER.key(), '{}')
        consumer: Location = codec.decode_string(cstr, Type[Location])
        pstr = attachments.get(Metadata.MESH_PROVIDER.key(), '{}')
        provider: Location = codec.decode_string(pstr, Type[Location])

        # Refresh context
        context = MeshContext()
        context.trace_id = trace_id
        context.span_id = span_id
        context.timestamp = self.resolve_timestamp(timestamp)
        context.run_mode = self.resolve_run_mode(run_mode)
        context.urn = urn
        context.consumer = consumer
        context.provider = provider
        for key, value in attachments.items():
            context.attachments[key] = value

        self.rewrite_context(context)

    @staticmethod
    def resolve_timestamp(v: str) -> int:
        if tool.optional(v):
            return int(time.time() * 1000)
        try:
            return int(v)
        except Exception as e:
            log.error("Parse timestamp failed. ", e)
            return int(time.time() * 1000)

    @staticmethod
    def resolve_run_mode(v: str) -> int:
        if tool.optional(v):
            return RunMode.ROUTINE.value
        try:
            return int(v)
        except Exception as e:
            log.error("Parse run mode failed.", e)
            return RunMode.ROUTINE.value


class MeshKey(Key):

    def __init__(self, name: str, kind: Type[T]):
        super().__init__(name)
        self.name = name
        self.kind = kind

    def get_if_absent(self) -> T:
        pass

    def map(self, fn) -> Optional[Any]:
        pass

    def if_present(self, fn):
        pass

    def or_else(self, v: T) -> T:
        pass

    def is_present(self) -> bool:
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import threading
from collections import deque
from contextvars import ContextVar
from typing import Any, Deque, Type, Coroutine

from mesh.context import MeshContext, MeshKey
from mesh.macro import T
from mesh.prsim import Context, Key

__context__: ContextVar[Deque[Context]] = ContextVar[Deque[Context]]('mesh_context', default=deque())


class Mesh:
    # Mesh invoke timeout
    TIMEOUT: Key[int] = MeshKey("mesh.mpc.timeout", int)

    # Mesh invoke mpi name attributes.
    UNAME: Key[str] = MeshKey("mesh.mpc.uname", str)

    # Mesh mpc remote address.
    REMOTE: Key[str] = MeshKey("mesh.mpc.address", str)

    # Remote app name.
    REMOTE_NAME: Key[str] = MeshKey("mesh.mpc.remote.name", str)

    # Generic codec type
    GENERIC_CODEC_TYPE: Key[str] = MeshKey("mesh.codec.generic.type", Type[T])

    @staticmethod
    def context() -> Context:
        mtx: Deque[Context] = __context__.get()
        if mtx.__len__() < 1:
            return MeshContext.create()
        return mtx[-1]

    @staticmethod
    def release():
        if not Mesh.is_empty():
            __context__.set(deque())

    @staticmethod
    def reset(ctx: Context):
        mtx = deque()
        mtx.append(ctx)
        __context__.set(mtx)

    @staticmethod
    def is_empty() -> bool:
        return __context__.get().__len__() < 1

    @staticmethod
    def push(ctx: Context):
        __context__.get().append(ctx)

    @staticmethod
    def pop():
        mtx: Deque[Context] = __context__.get()
        if not Mesh.is_empty():
            mtx.pop()

    @staticmethod
    async def context_safe(routine: Coroutine) -> Any:
        """ Execute with context safety. """
        context_disable = Mesh.is_empty()
        try:
            if context_disable:
                Mesh.reset(MeshContext.create())
            else:
                Mesh.push(Mesh.context().resume())
            return await routine
        finally:
            if context_disable:
                Mesh.release()
            else:
                Mesh.pop()

    @staticmethod
    def context_id() -> str:
        return f"{threading.current_thread().ident}-{threading.current_thread().name}"



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import Generic, Any, Dict, List

from mesh.kinds import Script, Paging, Page
from mesh.macro import mpi, spi, T, A


@spi("mesh")
class Evaluator(ABC, Generic[T, A]):
    """
    Evaluate engine.
    """

    @abstractmethod
    @mpi("mesh.eval.compile")
    async def compile(self, script: Script) -> str:
        """
        Compile the named rule.
        :param script:
        :return:
        """

    @abstractmethod
    @mpi("mesh.eval.exec")
    async def exec(self, code: str, args: Any, dft: str) -> str:
        """
        Exec the script with name.
        :param code:
        :param args:
        :param dft:
        :return:
        """

    @abstractmethod
    @mpi("mesh.eval.dump")
    async def dump(self, feature: Dict[str, str]) -> List[Script]:
        """
        Dump the scripts.
        :param feature:
        :return:
        """

    @abstractmethod
    @mpi("mesh.eval.index")
    async def index(self, index: Paging) -> Page[Script]:
        """
        Dump the scripts.
        :param index:
        :return:
        """



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import asyncio
import multiprocessing
import os
import signal
from asyncio import CancelledError
from multiprocessing import Process
from threading import Event
from typing import List

import mesh.environ as env
import mesh.log as log
import mesh.tool as tool
from mesh.macro import ServiceLoader
from mesh.mpc import Provider
from mesh.prsim import RuntimeHook


class MooterThread(RuntimeHook):
    def __init__(self):
        self.http1: Provider = ServiceLoader.load(Provider).get("http")
        self.http2: Provider = ServiceLoader.load(Provider).get("grpc")
        self.hooks = ServiceLoader.load(RuntimeHook).list('')
        self.waiter = Event()

    async def start(self):
        for hook in self.hooks:
            await hook.start()
        self.add_exit_hook()
        await self.http1.start(f"0.0.0.0:{tool.get_mesh_runtime().port + 1}", None)
        await self.http2.start(f"0.0.0.0:{tool.get_mesh_runtime().port}", None)

    async def stop(self):
        await self.http1.close()
        await self.http2.close()
        for hook in self.hooks:
            await hook.stop()

    async def refresh(self):
        await self.stop()
        await self.start()

    async def wait(self):
        try:
            await self.http1.wait()
            await self.http2.wait()
            await self.stop()
            # self.waiter.wait()
        except CancelledError as e:
            await self.stop()
            log.info(f"{env.System.environ().get_mesh_name()} has been stop gracefully. ")
        except KeyboardInterrupt:
            await self.stop()
            log.info(f"{env.System.environ().get_mesh_name()} has been stop gracefully. ")

    async def start_wait(self):
        await self.start()
        await self.wait()

    def add_exit_hook(self):
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

    def signal_handler(self, signum, frame):
        try:
            asyncio.run(self.stop())
        except KeyboardInterrupt:
            pass
        finally:
            self.waiter.set()


class Mooter(RuntimeHook):

    def __init__(self):
        self.main: RuntimeHook = MooterThread()
        self.workers: List[Process] = []

    @staticmethod
    def start_in_proc(addr: str):
        os.environ.setdefault("MESH_RUNTIME", addr)
        mooter = MooterThread()
        asyncio.run(mooter.start_wait())

    async def start(self):
        await self.main.start()
        self.workers.clear()
        for idx in range(min(tool.get_proc(), multiprocessing.cpu_count() * 2)):
            addr = f"0.0.0.0:{tool.get_mesh_runtime().port - 2 * (idx + 1)}"
            name = f"mesh-worker-{idx}"
            self.workers.append(Process(target=self.start_in_proc, name=name, args=(addr,)))
        for worker in self.workers:
            try:
                worker.start()
            except BaseException as e:
                log.error("", e)

    async def stop(self):
        await self.main.stop()
        for worker in self.workers:
            worker.terminate()
            worker.join()
        log.info("Mesh has been graceful exit.")

    async def refresh(self):
        await self.stop()
        await self.start()

    async def wait(self):
        await self.main.wait()
        await self.stop()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import y_pb2 as y__pb2


class PrivateTransferTransportStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.peek = channel.unary_unary(
                '/org.ppc.ptp.PrivateTransferTransport/peek',
                request_serializer=y__pb2.PeekInbound.SerializeToString,
                response_deserializer=y__pb2.TransportOutbound.FromString,
                )
        self.pop = channel.unary_unary(
                '/org.ppc.ptp.PrivateTransferTransport/pop',
                request_serializer=y__pb2.PopInbound.SerializeToString,
                response_deserializer=y__pb2.TransportOutbound.FromString,
                )
        self.push = channel.unary_unary(
                '/org.ppc.ptp.PrivateTransferTransport/push',
                request_serializer=y__pb2.PushInbound.SerializeToString,
                response_deserializer=y__pb2.TransportOutbound.FromString,
                )
        self.release = channel.unary_unary(
                '/org.ppc.ptp.PrivateTransferTransport/release',
                request_serializer=y__pb2.ReleaseInbound.SerializeToString,
                response_deserializer=y__pb2.TransportOutbound.FromString,
                )


class PrivateTransferTransportServicer(object):
    """Missing associated documentation comment in .proto file."""

    def peek(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def pop(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def push(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def release(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PrivateTransferTransportServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'peek': grpc.unary_unary_rpc_method_handler(
                    servicer.peek,
                    request_deserializer=y__pb2.PeekInbound.FromString,
                    response_serializer=y__pb2.TransportOutbound.SerializeToString,
            ),
            'pop': grpc.unary_unary_rpc_method_handler(
                    servicer.pop,
                    request_deserializer=y__pb2.PopInbound.FromString,
                    response_serializer=y__pb2.TransportOutbound.SerializeToString,
            ),
            'push': grpc.unary_unary_rpc_method_handler(
                    servicer.push,
                    request_deserializer=y__pb2.PushInbound.FromString,
                    response_serializer=y__pb2.TransportOutbound.SerializeToString,
            ),
            'release': grpc.unary_unary_rpc_method_handler(
                    servicer.release,
                    request_deserializer=y__pb2.ReleaseInbound.FromString,
                    response_serializer=y__pb2.TransportOutbound.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'org.ppc.ptp.PrivateTransferTransport', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class PrivateTransferTransport(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def peek(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/org.ppc.ptp.PrivateTransferTransport/peek',
            y__pb2.PeekInbound.SerializeToString,
            y__pb2.TransportOutbound.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def pop(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/org.ppc.ptp.PrivateTransferTransport/pop',
            y__pb2.PopInbound.SerializeToString,
            y__pb2.TransportOutbound.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def push(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/org.ppc.ptp.PrivateTransferTransport/push',
            y__pb2.PushInbound.SerializeToString,
            y__pb2.TransportOutbound.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def release(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/org.ppc.ptp.PrivateTransferTransport/release',
            y__pb2.ReleaseInbound.SerializeToString,
            y__pb2.TransportOutbound.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.kinds import Topic, Timeout
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Scheduler


@spi("mesh")
class MeshScheduler(Scheduler):

    def __init__(self):
        self.remote = ServiceProxy.default_proxy(Scheduler)

    async def timeout(self, timeout: Timeout, duration: int) -> str:
        return await self.remote.timeout(timeout, duration)

    async def cron(self, cron: str, binding: Topic) -> str:
        return await self.remote.cron(cron, binding)

    async def period(self, duration: int, binding: Topic) -> str:
        return await self.remote.period(duration, binding)

    async def dump(self) -> List[str]:
        return await self.remote.dump()

    async def cancel(self, task_id: str) -> bool:
        return await self.remote.cancel(task_id)

    async def stop(self, task_id: str) -> bool:
        return await self.remote.stop(task_id)

    async def emit(self, topic: Topic) -> bool:
        return await self.remote.emit(topic)

    async def shutdown(self, duration: int) -> bool:
        return await self.remote.shutdown(duration)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import asyncio
import traceback
from asyncio import Task
from typing import List, Dict, Callable, Coroutine

import mesh.log as log
import mesh.tool as tool
from mesh.kinds import Topic, Timeout, Event
from mesh.macro import spi, ServiceLoader, Binding
from mesh.mpc import ServiceProxy
from mesh.prsim import Scheduler, Subscriber


async def period_emitter(duration: int, trigger: Callable[[Topic], Coroutine], topic: Topic):
    while True:
        await trigger(topic)
        await asyncio.sleep(duration / 1000)


@spi("python")
class PythonScheduler(Scheduler):

    def __init__(self):
        self.remote = ServiceProxy.default_proxy(Scheduler)
        self.tasks: Dict[str, Task] = {}

    async def timeout(self, timeout: Timeout, duration: int) -> str:
        return ''

    async def cron(self, cron: str, binding: Topic) -> str:
        return ''

    async def period(self, duration: int, binding: Topic) -> str:
        task_id = tool.next_id()
        log.info(f"Next task {task_id} has been submit.")
        task = asyncio.create_task(period_emitter(duration, self.emit, binding))
        self.tasks[task_id] = task
        return task_id

    async def dump(self) -> List[str]:
        task_ids = []
        for task_id, _ in self.tasks.items():
            task_ids.append(task_id)
        return task_ids

    async def cancel(self, task_id: str) -> bool:
        return await self.stop(task_id)

    async def stop(self, task_id: str) -> bool:
        task = self.tasks.get(task_id, None)
        if task:
            task.cancel()
            self.tasks.__delitem__(task_id)
        return True

    async def emit(self, topic: Topic) -> bool:
        subscribers = ServiceLoader.load(Subscriber).list('')
        for subscriber in subscribers:
            try:
                bindings = Binding.get_binding_if_present(subscriber)
                if not topic.matches(bindings):
                    continue
                await subscriber.subscribe(await Event.new_instance({}, topic))
            except BaseException as e:
                print(traceback.format_exc())
                log.error(f"{e}")

        return True

    async def shutdown(self, duration: int) -> bool:
        for _, task in self.tasks.items():
            task.cancel()
        self.tasks.clear()
        return True



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import List

from mesh.kinds import Timeout, Topic
from mesh.macro import spi, mpi


@spi("python")
class Scheduler(ABC):

    @abstractmethod
    @mpi("mesh.schedule.timeout")
    async def timeout(self, timeout: Timeout, duration: int) -> str:
        """
        Schedules the specified {@link Timeout} for one-time execution after
        the specified delay.
        :param timeout:
        :param duration:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.cron")
    async def cron(self, cron: str, binding: Topic) -> str:
        """
        Schedules with the cron expression. "0 * * 1-3 * ? *"
        :param cron:
        :param binding:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.period")
    async def period(self, duration: int, binding: Topic) -> str:
        """
        Releases all resources acquired by this {@link Scheduler} and cancels all
        tasks which were scheduled but not executed yet.
        :param duration:
        :param binding:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.dump")
    async def dump(self) -> List[str]:
        """
        Dump all taskIds, only effected in pvm.
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.cancel")
    async def cancel(self, task_id: str) -> bool:
        """
        Cancel the pending tasks.
        :param task_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.stop")
    async def stop(self, task_id: str) -> bool:
        """
        Stop the pending tasks.
        :param task_id:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.emit")
    async def emit(self, binding: Topic) -> bool:
        """
        Shutdown the scheduler.
        :param binding:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.schedule.shutdown")
    async def shutdown(self, duration: int) -> bool:
        """
        Shutdown the scheduler.
        :param duration:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2023, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any

import grpc

import mesh.codec.tools as tools
from mesh.context import Mesh
from mesh.grpx.channels import GrpcChannels
from mesh.kinds import Reference
from mesh.mpc.invoker import Execution
from mesh.ptp.y_pb2 import PeekInbound, PopInbound, PushInbound, ReleaseInbound, TransportOutbound


class BfiaBinding:
    def __init__(self, urn: str, args: Any):
        if urn.startswith("peek.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/peek"
            self.inbound = PeekInbound()
            self.inbound.topic = args.topic
            self.outbound = TransportOutbound
        elif urn.startswith("pop.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/pop"
            self.inbound = PopInbound()
            self.inbound.topic = args.topic
            self.inbound.timeout = args.timeout
            self.outbound = TransportOutbound
        elif urn.startswith("push.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/push"
            self.inbound = PushInbound()
            self.inbound.topic = args.topic
            self.inbound.payload = args.payload
            if args.metadata:
                for k, v in args.metadata.items():
                    self.inbound.metadata[k] = v
            self.outbound = TransportOutbound
        elif urn.startswith("release.chan.mesh"):
            self.path = "/org.ppc.ptp.PrivateTransferTransport/release"
            self.inbound = ReleaseInbound()
            self.inbound.topic = args.topic
            self.inbound.timeout = args.timeout
            self.outbound = TransportOutbound
        else:
            self.path = ""
            self.inbound = None
            self.outbound = None

    def encode(self, _: bytes):
        return self.inbound.SerializeToString()

    def decode(self, buffer: bytes):
        out = self.outbound()
        out.ParseFromString(buffer)
        v = "{" + f'"code":"{out.code}","message":"{out.message}","content":"{tools.b64encode(out.payload)}"' + "}"
        return v.encode('utf-8')


async def bfia_consume(address: str, binding: BfiaBinding, execution: Execution[Reference], inbound: bytes,
                       channel: GrpcChannels) -> bytes:
    urn = Mesh.context().get_urn()
    try:
        Mesh.context().rewrite_urn("")
        ct = Mesh.context().get_attribute(Mesh.TIMEOUT)
        timeout = (ct if ct else execution.schema().timeout) / 1000
        channel = channel.create_if_absent(address)
        conn = channel.unary_unary(binding.path, binding.encode, binding.decode)
        return await conn(inbound, timeout=timeout, metadata=[], wait_for_ready=True, compression=grpc.Compression.Gzip)
    finally:
        Mesh.context().rewrite_urn(urn)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: x.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='x.proto',
  package='org.ppc.ptp',
  syntax='proto3',
  serialized_options=b'\n\013io.bfia.ptpZ+gitlab.trustbe.net/middleware/gaia-mesh/ptp',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x07x.proto\x12\x0borg.ppc.ptp\"\x81\x01\n\x07Inbound\x12\x34\n\x08metadata\x18\x01 \x03(\x0b\x32\".org.ppc.ptp.Inbound.MetadataEntry\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa2\x01\n\x08Outbound\x12\x35\n\x08metadata\x18\x01 \x03(\x0b\x32#.org.ppc.ptp.Outbound.MetadataEntry\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x12\x0c\n\x04\x63ode\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32\x8e\x01\n\x17PrivateTransferProtocol\x12<\n\ttransport\x12\x14.org.ppc.ptp.Inbound\x1a\x15.org.ppc.ptp.Outbound(\x01\x30\x01\x12\x35\n\x06invoke\x12\x14.org.ppc.ptp.Inbound\x1a\x15.org.ppc.ptp.OutboundB:\n\x0bio.bfia.ptpZ+gitlab.trustbe.net/middleware/gaia-mesh/ptpb\x06proto3'
)




_INBOUND_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='org.ppc.ptp.Inbound.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='org.ppc.ptp.Inbound.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='org.ppc.ptp.Inbound.MetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=154,
)

_INBOUND = _descriptor.Descriptor(
  name='Inbound',
  full_name='org.ppc.ptp.Inbound',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='org.ppc.ptp.Inbound.metadata', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='org.ppc.ptp.Inbound.payload', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_INBOUND_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=154,
)


_OUTBOUND_METADATAENTRY = _descriptor.Descriptor(
  name='MetadataEntry',
  full_name='org.ppc.ptp.Outbound.MetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='org.ppc.ptp.Outbound.MetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='org.ppc.ptp.Outbound.MetadataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=154,
)

_OUTBOUND = _descriptor.Descriptor(
  name='Outbound',
  full_name='org.ppc.ptp.Outbound',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='org.ppc.ptp.Outbound.metadata', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='org.ppc.ptp.Outbound.payload', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='org.ppc.ptp.Outbound.code', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='org.ppc.ptp.Outbound.message', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_OUTBOUND_METADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=157,
  serialized_end=319,
)

_INBOUND_METADATAENTRY.containing_type = _INBOUND
_INBOUND.fields_by_name['metadata'].message_type = _INBOUND_METADATAENTRY
_OUTBOUND_METADATAENTRY.containing_type = _OUTBOUND
_OUTBOUND.fields_by_name['metadata'].message_type = _OUTBOUND_METADATAENTRY
DESCRIPTOR.message_types_by_name['Inbound'] = _INBOUND
DESCRIPTOR.message_types_by_name['Outbound'] = _OUTBOUND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Inbound = _reflection.GeneratedProtocolMessageType('Inbound', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _INBOUND_METADATAENTRY,
    '__module__' : 'x_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.Inbound.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _INBOUND,
  '__module__' : 'x_pb2'
  # @@protoc_insertion_point(class_scope:org.ppc.ptp.Inbound)
  })
_sym_db.RegisterMessage(Inbound)
_sym_db.RegisterMessage(Inbound.MetadataEntry)

Outbound = _reflection.GeneratedProtocolMessageType('Outbound', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _OUTBOUND_METADATAENTRY,
    '__module__' : 'x_pb2'
    # @@protoc_insertion_point(class_scope:org.ppc.ptp.Outbound.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _OUTBOUND,
  '__module__' : 'x_pb2'
  # @@protoc_insertion_point(class_scope:org.ppc.ptp.Outbound)
  })
_sym_db.RegisterMessage(Outbound)
_sym_db.RegisterMessage(Outbound.MetadataEntry)


DESCRIPTOR._options = None
_INBOUND_METADATAENTRY._options = None
_OUTBOUND_METADATAENTRY._options = None

_PRIVATETRANSFERPROTOCOL = _descriptor.ServiceDescriptor(
  name='PrivateTransferProtocol',
  full_name='org.ppc.ptp.PrivateTransferProtocol',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=322,
  serialized_end=464,
  methods=[
  _descriptor.MethodDescriptor(
    name='transport',
    full_name='org.ppc.ptp.PrivateTransferProtocol.transport',
    index=0,
    containing_service=None,
    input_type=_INBOUND,
    output_type=_OUTBOUND,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='invoke',
    full_name='org.ppc.ptp.PrivateTransferProtocol.invoke',
    index=1,
    containing_service=None,
    input_type=_INBOUND,
    output_type=_OUTBOUND,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_PRIVATETRANSFERPROTOCOL)

DESCRIPTOR.services_by_name['PrivateTransferProtocol'] = _PRIVATETRANSFERPROTOCOL

# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from asyncio import CancelledError
from typing import Any

import grpc
from grpc.aio import Server

import mesh.log as log
import mesh.tool as tool
from mesh.cause import ValidationException
from mesh.grpx import GrpcBindableService, AsyncGrpcInterceptor
from mesh.macro import spi
from mesh.mpc import Provider


@spi("grpc")
class GrpcProvider(Provider):
    """
    Grpc provider wrap the grpc protocol implementation.
    """

    def __init__(self):
        self.interceptor = AsyncGrpcInterceptor()
        self.service = GrpcBindableService()
        self.address = ""
        self.server: Server = None

    async def start(self, address: str, tc: Any):
        self.address = address
        if not self.server:
            self.server = grpc.aio.server(
                handlers=[self.service],
                interceptors=[self.interceptor],
                options=[('grpc.max_message_length', 1 << 30),
                         ('grpc.max_send_message_length', 1 << 30),
                         ('grpc.max_receive_message_length', 1 << 30)],
                compression=grpc.Compression.Gzip,
            )
        if tool.optional(address):
            raise ValidationException("GRPC address cant be empty.")
        self.server.add_insecure_port(address)
        self.server.add_generic_rpc_handlers([self.service])
        await self.server.start()
        log.info(f"Listening and serving HTTP 2.0 on {address}")

    async def close(self):
        if not self.server:
            return
        try:
            await self.server.stop(grace=12)
        except CancelledError as e:
            pass
        finally:
            log.info(f"Graceful stop HTTP 2.0 serving on {self.address}")

    async def wait(self):
        if not self.server:
            return
        try:
            await self.server.wait_for_termination()
        except CancelledError as e:
            pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod
from typing import Any

from mesh.macro import spi, T, ServiceLoader
from mesh.mpc.invoker import Invocation, Invoker

PROVIDER = "PROVIDER"
CONSUMER = "CONSUMER"


@spi(pattern="*")
class Filter:

    @abstractmethod
    async def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        pass

    @staticmethod
    def composite(invoker: Invoker[T], pattern: str) -> Invoker[T]:
        """
        Composite the filter spi providers as a invoker.
        :param invoker: Delegate invoker.
        :param pattern: Filter pattern.
        :return: Composite invoker.
        """
        filters = ServiceLoader.load(Filter).list(pattern)
        last: Invoker[T] = invoker
        for index, filter_ in enumerate(filters):
            next_: Invoker[T] = last
            last = FilteredInvoker(next_, filter_)
        return last


class FilteredInvoker(Invoker[T]):

    def __init__(self, next_: Invoker[T], filter_: Filter):
        self.next_ = next_
        self.filter_ = filter_

    async def run(self, invocation: Invocation) -> Any:
        return await self.filter_.invoke(self.next_, invocation)


async def invoke(invoker: Invoker[T], filtering: Filter, invocation: Invocation):
    return await filtering.invoke(invoker, invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod
from enum import Enum


class Codeable:
    """
    /**
     * See https://blue-elephant.yuque.com/blue-elephant/dg3hry/srnhhr.
     *
     * <pre>
     *
     * 1.5版本后，算子、算法、业务应用均往该规范靠.
     *
     * 错误码格式：
     *     E      0       00       000         0000
     *    前缀  异常等级  异常类型  模块或应用    业务码含义
     *
     * 总共10位错误码:
     * 1. 1位是等级，为0的等级直接对客
     * 2. 23位是异常类型
     * 3. 456位是模块或应用标志
     * 4. 78910位是自定义的业务码含义.
     *
     * 错误message可以是开发便于排查的，在网关层转化为产品定义的message，可配置.
     *
     * 错误码	模块
     * 000	   mesh
     * 001	   omega
     * 002	   asset
     * 003	   theta
     * 004	   cube
     * 005	   edge
     * 006     tensor
     * 007     base
     * </pre>
     *
     * @author coyzeng@gmail.com
     */
    """

    @abstractmethod
    def get_code(self) -> str:
        pass

    @abstractmethod
    def get_message(self) -> str:
        pass

    def match(self, codeable: "Codeable") -> bool:
        """
        Error code is matchable.
        """
        return self.matches(codeable.get_code())

    def matches(self, code: str) -> bool:
        if code is None:
            return False
        return str.__eq__(self.get_code(), code)


class MeshCode(Codeable, Enum):
    SUCCESS = "E0000000000", "请求成功"
    NOT_FOUND = "E0000000404", "请求资源不存在"
    SYSTEM_ERROR = "E0000000500", "系统异常"
    SERVICE_UNAVAILABLE = "E0000000503", "循环请求服务不可达"
    VALIDATE_ERROR = "E0000000400", "请求非法"
    UNAUTHORIZED = "E0000000403", "请求资源未被授权"
    UNKNOWN = "E0000000520", "未知异常"
    COMPATIBLE_ERROR = "E0000000600", "系统不兼容"
    TIMEOUT_ERROR = "E0000000601", "请求超时"
    NO_PROVIDER_ERROR = "E0000000602", "无服务实例"
    CRYPT_ERROR = "E0000000603", "数字证书校验异常"
    TOKEN_EXPIRE = "E0000000604", "节点授权码已过期"
    NET_EXPIRE = "E0000000605", "节点组网时间已过期"
    NET_DISABLE = "E0000000606", "对方节点已禁用网络"
    NET_UNAVAILABLE = "E0000000607", "网络不通"
    LICENSE_FORMAT_ERROR = "E0000000608", "License非法"
    LICENSE_EXPIRED = "E0000000609", "License已过期"
    MAX_REPLICAS_LIMIT = "E0000000610", "集群副本已达许可上限"
    MAX_COOP_LIMIT = "E0000000611", "合作方链接数已达许可上限"
    MAX_TENANT_LIMIT = "E0000000612", "租户数已达许可上限"
    MAX_USER_LIMIT = "E0000000613", "用户数已达许可上限"
    URN_NOT_PERMIT = "E0000000614", "接口未被许可调用"
    SIGNATURE_ERROR = "E0000000615", "证书签名非法"
    CRYPT_CODEC_ERROR = "E0000000616", "报文编解码异常"
    NO_SERVICE = "E0000000617", "下游版本不匹配服务不存在"
    NET_NOT_WEAVE = "E0000000618", "节点或机构未组网"
    ADDRESS_ERROR = "E0000000619", "地址非法或无法访问"
    UNEXPECTED_SYNTAX = "E0000000620", "规则语法错误"
    UNKNOWN_SCRIPT = "E0000000621", "规则不存在"
    DATA_ID_ERROR = "E0000000622", "数据身份不存在或非法"
    START_PENDING = "E0000000623", "系统启动中"
    UNEXPECTED_SQL_SYNTAX = "E0000000624", "SQL语法错误"
    NO_DATABASE = "E0000000625", "数据库不存在"
    NO_TRANSACTION = "E0000000626", "事务不存在"
    SQL_EXECUTE_ERROR = "E0000000627", "SQL执行异常"
    DUPLICATE_KEY = "E0000000628", "唯一键冲突"

    def __init__(self, code: str, message: str) -> None:
        self.code = code
        self.message = message

    def __str__(self):
        return f'{self.code}({self.message})'

    def get_code(self) -> str:
        return self.value[0]

    def get_message(self) -> str:
        return self.value[1]



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.kinds.principal import Principal
from mesh.macro import index, serializable


@serializable
class Location(Principal):

    @staticmethod
    async def localize(self: "Location"):
        from mesh.prsim import Network
        from mesh.macro import ServiceLoader
        import mesh.tool as tool
        network = ServiceLoader.load(Network).get_default()
        environ = await network.get_environ()
        self.inst_id = environ.inst_id
        self.node_id = environ.node_id
        self.ip = tool.get_mesh_direct()
        self.host = tool.get_hostname()
        self.port = f"{tool.get_mesh_runtime().port}"
        self.name = tool.get_mesh_name()

    @index(10)
    def ip(self) -> str:
        return ""

    @index(15)
    def port(self) -> str:
        return ""

    @index(20)
    def host(self) -> str:
        return ""

    @index(25)
    def name(self) -> str:
        return ""



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod
from typing import Dict

from mesh.kinds import License
from mesh.macro import mpi


class Licenser(ABC):

    @abstractmethod
    @mpi(name="mesh.license.imports", flags=2)
    async def imports(self, text: str):
        """
        Import the license.
        :param text:
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.license.exports", flags=2)
    async def exports(self) -> str:
        """
        Exports the license.
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.license.explain", flags=2)
    async def explain(self) -> License:
        """
        Explain the license.
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.license.verify", flags=2)
    async def verify(self) -> bool:
        """
        Verify the license.
        :return:
        """
        pass

    @abstractmethod
    @mpi(name="mesh.license.features", flags=2)
    async def features(self) -> Dict[str, str]:
        """
        License features.
        :return:
        """
        pass



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import x_pb2 as x__pb2


class PrivateTransferProtocolStub(object):
    """互联互通如果使用异步传输协议作为标准参考，Header会复用metadata传输互联互通协议报头，且metadata中会传输异步场景下的消息相关属性
    互联互通如果使用其他协议作为参考标准，Header会复用metadata传输互联互通协议报头
    互联互通如果使用GRPC作为参考标准，Header会复用HTTP2的报头传输互联互通协议报头

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.transport = channel.stream_stream(
                '/org.ppc.ptp.PrivateTransferProtocol/transport',
                request_serializer=x__pb2.Inbound.SerializeToString,
                response_deserializer=x__pb2.Outbound.FromString,
                )
        self.invoke = channel.unary_unary(
                '/org.ppc.ptp.PrivateTransferProtocol/invoke',
                request_serializer=x__pb2.Inbound.SerializeToString,
                response_deserializer=x__pb2.Outbound.FromString,
                )


class PrivateTransferProtocolServicer(object):
    """互联互通如果使用异步传输协议作为标准参考，Header会复用metadata传输互联互通协议报头，且metadata中会传输异步场景下的消息相关属性
    互联互通如果使用其他协议作为参考标准，Header会复用metadata传输互联互通协议报头
    互联互通如果使用GRPC作为参考标准，Header会复用HTTP2的报头传输互联互通协议报头

    """

    def transport(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def invoke(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PrivateTransferProtocolServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'transport': grpc.stream_stream_rpc_method_handler(
                    servicer.transport,
                    request_deserializer=x__pb2.Inbound.FromString,
                    response_serializer=x__pb2.Outbound.SerializeToString,
            ),
            'invoke': grpc.unary_unary_rpc_method_handler(
                    servicer.invoke,
                    request_deserializer=x__pb2.Inbound.FromString,
                    response_serializer=x__pb2.Outbound.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'org.ppc.ptp.PrivateTransferProtocol', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class PrivateTransferProtocol(object):
    """互联互通如果使用异步传输协议作为标准参考，Header会复用metadata传输互联互通协议报头，且metadata中会传输异步场景下的消息相关属性
    互联互通如果使用其他协议作为参考标准，Header会复用metadata传输互联互通协议报头
    互联互通如果使用GRPC作为参考标准，Header会复用HTTP2的报头传输互联互通协议报头

    """

    @staticmethod
    def transport(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/org.ppc.ptp.PrivateTransferProtocol/transport',
            x__pb2.Inbound.SerializeToString,
            x__pb2.Outbound.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def invoke(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/org.ppc.ptp.PrivateTransferProtocol/invoke',
            x__pb2.Inbound.SerializeToString,
            x__pb2.Outbound.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, Dict

from mesh.cause import MeshCode, Codeable
from mesh.codec import Codec
from mesh.codec import Json
from mesh.context import Mesh
from mesh.kinds import Location
from mesh.macro import spi, ServiceLoader
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, CONSUMER
from mesh.prsim import Metadata


@spi(name="consumer", pattern=CONSUMER, priority=1 - (1 << 32))
class ConsumerFilter(Filter):

    async def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        codec = ServiceLoader.load(Codec).get(Json)
        consumer = Mesh.context().get_consumer()
        if consumer.node_id == "":
            await Location.localize(consumer)
        provider = Mesh.context().get_provider()
        if provider.node_id == "":
            await Location.localize(provider)

        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        for key, value in Mesh.context().get_attachments().items():
            attachments[key] = value
        attachments[Metadata.MESH_TRACE_ID.key()] = Mesh.context().get_trace_id()
        attachments[Metadata.MESH_SPAN_ID.key()] = Mesh.context().get_span_id()
        attachments[Metadata.MESH_TIMESTAMP.key()] = f"{Mesh.context().get_timestamp()}"
        attachments[Metadata.MESH_RUN_MODE.key()] = f"{Mesh.context().get_run_mode()}"
        attachments[Metadata.MESH_URN.key()] = Mesh.context().get_urn()
        attachments[Metadata.MESH_CONSUMER.key()] = codec.encode_string(consumer)
        attachments[Metadata.MESH_PROVIDER.key()] = codec.encode_string(provider)

        digest = Digest()
        try:
            ret = await invoker.run(invocation)
            digest.write("C", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("C", e.get_code())
            else:
                digest.write("C", MeshCode.SYSTEM_ERROR.get_code())
            raise e



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod
from typing import Any

from mesh.macro import spi


@spi(name="grpc")
class Provider:
    """

    """

    @abstractmethod
    async def start(self, address: str, tc: Any):
        """
        Start the mesh broker.
        :param address:
        :param tc:
        :return:
        """
        pass

    @abstractmethod
    async def close(self):
        pass

    @abstractmethod
    async def wait(self):
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from asyncio import CancelledError
from concurrent.futures import Future
from typing import Iterator, Union, Any

import grpc

import mesh.tool as tool
from mesh.cause import TimeoutException, MeshException, NotFoundException, MeshCode
from mesh.context import Mesh
from mesh.environ import System
from mesh.grpx.bfia import BfiaBinding, bfia_consume
from mesh.grpx.channels import GrpcChannels
from mesh.kinds import Reference
from mesh.macro import spi
from mesh.mpc.consumer import Consumer
from mesh.mpc.invoker import Execution
from mesh.mpc.urn import URN


@spi("grpc")
class GrpcConsumer(Consumer):

    def __init__(self):
        self.channel = GrpcChannels()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Close """
        self.channel.__exit__(exc_type, exc_val, exc_tb)

    def start(self):
        """ Do not need to implement """
        pass

    async def consume(self, address: str, urn: URN, execution: Execution[Reference], inbound: bytes,
                      args: Any) -> bytes:
        """
        request_iterator=iterator,
        timeout=None,
        metadata=None,
        credentials=None,
        wait_for_ready=None,
        compression=None
        :param address:
        :param urn:
        :param execution:
        :param inbound:
        :param args:
        :return:
        """
        binding = BfiaBinding(urn.string(), args)
        if "" != binding.path:
            return await bfia_consume(address, binding, execution, inbound, self.channel)

        ct = Mesh.context().get_attribute(Mesh.TIMEOUT)
        timeout = (ct if ct else execution.schema().timeout) / 1000
        try:
            return await self.channel.unary(address, inbound, timeout, [])
        except grpc.FutureTimeoutError:
            raise TimeoutException(f'Invoke {execution.schema().urn} timeout with {timeout * 1000}ms')
        except grpc.FutureCancelledError:
            raise MeshException(MeshCode.SYSTEM_ERROR.get_code(), f'Invoke {execution.schema().urn} canceled')
        except CancelledError as e:
            raise e
        except grpc.RpcError as e:
            if isinstance(e, grpc.Call):
                if e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                    raise TimeoutException(f'Invoke {urn.name} timeout with {execution.schema().timeout}ms')
                if e.code() == grpc.StatusCode.UNKNOWN:
                    raise NotFoundException(f'Service {urn.name} not found')
                raise MeshException(MeshCode.SYSTEM_ERROR.get_code(), f'Invoke {urn.name} {e.code()} {e.details()}')
            raise e

    @staticmethod
    def resolve(address: str):
        hosts = tool.split(address, ":")
        if hosts.__len__() < 2:
            return f'{hosts[0]}:{System.environ().get_default_mesh_port()}'
        return address

    def reset_if_unconnected(self, channel: GrpcChannels):
        def x(e: BaseException):
            pass

        return x


class AsyncInvokeObserver(Future, grpc.Call):

    def __init__(self, rendezvous: Union[grpc.Future, grpc.Call]):
        super().__init__()
        self.rendezvous = rendezvous

    def initial_metadata(self):
        self.rendezvous.initial_metadata()

    def trailing_metadata(self):
        return self.rendezvous.trailing_metadata()

    def code(self):
        return self.rendezvous.code()

    def details(self):
        return self.rendezvous.details()

    def is_active(self):
        return self.rendezvous.is_active()

    def time_remaining(self):
        return self.rendezvous.time_remaining()

    def add_callback(self, callback):
        return self.rendezvous.add_callback(callback)


class IterableBuffer(Iterator):

    def __init__(self, buffer: bytes):
        self.buffer = list()
        self.buffer.append(buffer)

    def __next__(self) -> bytes:
        if self.buffer:
            return self.buffer.pop()
        raise StopIteration()


class GrpcFuture(grpc.Future):

    def __init__(self, future: grpc.Future, hooks: Any) -> None:
        self.future = future
        self.hooks = hooks

    def cancel(self, msg: Any = ...) -> bool:
        return self.future.cancel()

    def cancelled(self):
        return self.future.cancelled()

    def running(self):
        return self.future.running()

    def done(self):
        return self.future.done()

    def result(self, timeout=None):
        return self.future.result(timeout)

    def exception(self, timeout=None):
        return self.future.exception(timeout)

    def traceback(self, timeout=None):
        return self.future.traceback(timeout)

    def add_done_callback(self, fn):
        return self.future.add_done_callback(fn)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.kinds import Registration
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Registry


@spi("mesh")
class MeshRegistry(Registry):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Registry)

    async def register(self, registration: Registration):
        return await self.proxy.register(registration)

    async def registers(self, registrations: List[Registration]):
        return await self.proxy.registers(registrations)

    async def unregister(self, registration: Registration):
        return await self.proxy.unregister(registration)

    async def export(self, kind: str) -> List[Registration]:
        return await self.proxy.export()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod

import mesh.log as log
from mesh.cause import MeshCode, Codeable
from mesh.codec import Codec
from mesh.context import Mesh
from mesh.kinds import Outbound, MeshFlag, Service
from mesh.macro import spi, ServiceLoader, Cause
from mesh.mpc.eden import Eden
from mesh.mpc.invoker import ServiceInvocation, Execution
from mesh.mpc.urn import URN
from mesh.prsim import Context

PROVIDER = "provider"
CONSUMER = "consumer"


@spi(PROVIDER)
class Transporter(ABC):

    @abstractmethod
    async def transport(self, ctx: Context, urn: str, buff: bytes) -> bytes:
        """
        Transport the stream.
        :param ctx:
        :param urn:
        :param buff:
        :return:
        """
        pass

    @staticmethod
    def get_codec(flag: str) -> Codec:
        codec = ServiceLoader.load(Codec).get(MeshFlag.of_code(flag).get_name())
        if codec:
            return codec
        return ServiceLoader.load(Codec).get_default()


@spi(PROVIDER)
class ProviderTransporter(Transporter):

    async def transport(self, ctx: Context, urn: str, buff: bytes) -> bytes:
        return await Mesh.context_safe(self.transport_safe(ctx, urn, buff))

    async def transport_safe(self, ctx: Context, uname: str, buff: bytes) -> bytes:
        try:
            Mesh.context().rewrite_context(ctx)
            Mesh.context().rewrite_urn(uname)
            urn = URN.parse(uname)
            eden = ServiceLoader.load(Eden).get_default()
            codec = self.get_codec(urn.flag.codec)
            execution = eden.infer(uname)
            if execution is None:
                outbound = Outbound()
                outbound.code = MeshCode.NO_SERVICE.code
                outbound.message = f"No mpi named {urn.name}."
                return codec.encode(outbound)
            return await self.service(urn, codec, execution, buff)
        except BaseException as e:
            trace_id = Mesh.context().get_trace_id()
            span_id = Mesh.context().get_span_id()
            log.error(f"{trace_id}#{span_id} Invoke service {uname} with error.", e)
            outbound = Outbound()
            outbound.code = MeshCode.SYSTEM_ERROR.code
            outbound.message = str(e)
            outbound.cause = Cause.of(e)
            return ServiceLoader.load(Codec).get_default().encode(outbound)

    @staticmethod
    async def service(urn: URN, codec: Codec, execution: Execution[Service], buff: bytes) -> bytes:
        returns = execution.reflect()
        try:
            parameters = codec.decode(buff, execution.intype())
            invocation = ServiceInvocation(execution, execution.inspect(), parameters, execution, urn)
            result = await execution.run(invocation)
            returns.set_code(MeshCode.SUCCESS.get_code())
            returns.set_message(MeshCode.SUCCESS.get_message())
            returns.set_content(result)
            return codec.encode(returns)
        except BaseException as e:
            log.error("{}#{} Invoke service {} with error.",
                      Mesh.context().get_trace_id(), Mesh.context().get_trace_id(), Mesh.context().get_urn(), e)
            if isinstance(e, Codeable):
                returns.set_code(e.get_code())
                returns.set_message(e.get_message())
            else:
                returns.set_code(MeshCode.SYSTEM_ERROR.code)
                returns.set_message(str(e))
            returns.set_cause(Cause.of(e))
            return codec.encode(returns)


@spi(CONSUMER)
class ConsumerTransporter(Transporter):

    async def transport(self, ctx: Context, urn: str, buff: bytes) -> bytes:
        Mesh.context().rewrite_context(ctx)
        return b'unknown'



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.prsim.builtin import Builtin
from mesh.prsim.cache import Cache
from mesh.prsim.cluster import Cluster
from mesh.prsim.commerce import Commercialize
from mesh.prsim.context import Context, Key, Metadata, Queue, RunMode
from mesh.prsim.cryptor import Cryptor
from mesh.prsim.datahouse import DataHouse
from mesh.prsim.dispatcher import Dispatcher
from mesh.prsim.endpoint import Endpoint, EndpointSticker
from mesh.prsim.evaluator import Evaluator
from mesh.prsim.graphics import Graphics
from mesh.prsim.hodor import Hodor
from mesh.prsim.iostream import IOStream
from mesh.prsim.kv import KV
from mesh.prsim.licenser import Licenser
from mesh.prsim.locker import Locker
from mesh.prsim.network import Network
from mesh.prsim.publisher import Publisher
from mesh.prsim.registry import Registry
from mesh.prsim.routable import Routable
from mesh.prsim.runtime_hook import RuntimeHook
from mesh.prsim.scheduler import Scheduler
from mesh.prsim.sequence import Sequence
from mesh.prsim.subscriber import Subscriber
from mesh.prsim.tokenizer import Tokenizer
from mesh.prsim.transport import Transport, Session

__all__ = (
    "Builtin",
    "Cache",
    "Cluster",
    "Commercialize",
    "RunMode",
    "Key",
    "Metadata",
    "Queue",
    "Context",
    "Cryptor",
    "DataHouse",
    "Dispatcher",
    "Endpoint",
    "EndpointSticker",
    "Evaluator",
    "Graphics",
    "Hodor",
    "IOStream",
    "KV",
    "Licenser",
    "Locker",
    "Network",
    "Publisher",
    "Registry",
    "Routable",
    "RuntimeHook",
    "Scheduler",
    "Sequence",
    "Subscriber",
    "Tokenizer",
    "Transport",
    "Session",
)


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from queue import Queue
from typing import Dict

from grpc import StreamStreamMultiCallable

from mesh.grpx.channels import GrpcChannels
from mesh.macro import spi
from mesh.prsim import Transport, Session
from mesh.tool import get_mesh_address


@spi("grpc")
class GrpcTransport(Transport):

    def __init__(self):
        self.channel = GrpcChannels()
        self.sessions: Dict[str, Session] = {}

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(5000)
        self.channel.__exit__(exc_type, exc_val, exc_tb)

    def open(self, session_id: str, metadata: Dict[str, str]) -> Session:
        session = self.sessions.get(session_id, None)
        if not session:
            stream = self.channel.create_stream(get_mesh_address().any())
            session = self.sessions[session_id] = GrpcSession(stream)
        return session

    def close(self, timeout: int):
        for session_id, session in self.sessions.items():
            session.release(timeout)

    def roundtrip(self, payload: bytes, metadata: Dict[str, str]) -> bytes:
        """"""
        pass


@spi("grpc")
class GrpcSession(Session):

    def __init__(self, stream: StreamStreamMultiCallable):
        self.stream = stream
        self.inbounds = Queue()
        self.closed = False
        self.outbounds = self.stream(self.iter, 10000, [('urn', '')], None, True, False)

    def peek(self, topic: str = "") -> bytes:
        return self.outbounds.__next__()

    def pop(self, timeout: int, topic: str = "") -> bytes:
        return self.outbounds.__next__()

    def push(self, payload: bytes, metadata: Dict[str, str], topic: str = ""):
        self.inbounds.put(payload)

    def release(self, timeout: int, topic: str = ""):
        """"""
        pass

    def iter(self):
        while True:
            if self.closed:
                return
            payload = self.inbounds.get()
            if not payload:
                continue
            yield payload



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import ABC, abstractmethod

from mesh.macro import spi, mpi


@spi("mesh")
class Tokenizer(ABC):

    @abstractmethod
    @mpi("mesh.trust.apply")
    async def apply(self, kind: str, duration: int) -> str:
        """
        Apply a node token.
        :param kind:
        :param duration:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.trust.verify")
    async def verify(self, token: str) -> bool:
        """
        Verify some token verifiable.
        :param token:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import asyncio
from functools import wraps
from typing import Any


class CallableMember:

    def __init__(self, name: str, v: Any):
        self.name = name
        self.v = v

    def __getitem__(self, item):
        v = object.__getattribute__(self, 'v')
        if v is None:
            return None
        return v[item]

    def __call__(self, *args, **kwargs):
        return object.__getattribute__(self, 'v')

    def __getattribute__(self, name: str):
        if name == object.__getattribute__(self, 'name'):
            return self
        v = object.__getattribute__(self, 'v')
        if v is None:
            return None
        return object.__getattribute__(v, name)


class Cache(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    """  # noqa

    def __init__(self, fn):
        self.__doc__ = getattr(fn, "__doc__")
        self.fn = fn
        self.name = self.fn.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self

        if asyncio and asyncio.iscoroutinefunction(self.fn):
            return self._wrap_in_coroutine(obj)

        if not hasattr(self, self.name):
            setattr(self, self.name, self.fn(obj))

        return CallableMember(self.name, getattr(self, self.name))

    def _wrap_in_coroutine(self, obj):
        @wraps(obj)
        async def wrapper():
            future = asyncio.ensure_future(self.fn(obj))
            obj.__dict__[self.fn.__name__] = future
            return future

        return wrapper()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List

from mesh.context import MeshKey
from mesh.environ import Mode
from mesh.kinds import Route, Environ, Versions, Paging, Page, Institution
from mesh.macro import spi
from mesh.mpc import ServiceProxy, Mesh
from mesh.prsim import Network


@spi("mesh")
class MeshNetwork(Network):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Network)
        self.environ_key = MeshKey("mesh-environ", Environ)
        self.environ: Environ = Environ.default() if Mode.Isolate.enable() else None

    async def get_environ(self) -> Environ:
        if self.environ:
            return self.environ
        if Mesh.context().get_attribute(self.environ_key):
            return Mesh.context().get_attribute(self.environ_key)
        environ = Environ.default()
        Mesh.context().set_attribute(self.environ_key, environ)

        self.environ = await Mesh.context_safe(self.get_environ_safe())
        return self.environ

    async def get_environ_safe(self) -> Environ:
        Mesh.context().get_principals().clear()
        return await self.proxy.get_environ()

    async def accessible(self, route: Route) -> bool:
        return await self.proxy.accessible(route)

    async def refresh(self, routes: List[Route]):
        return await self.proxy.refresh(routes)

    async def get_route(self, node_id: str) -> Route:
        return await self.proxy.get_route(node_id)

    async def get_routes(self) -> List[Route]:
        return await self.proxy.get_routes()

    async def get_domains(self) -> List[Route]:
        return await self.proxy.get_domains()

    async def put_domains(self, domains: List[Route]):
        return await self.proxy.put_domains(domains)

    async def weave(self, route: Route) -> None:
        return await self.proxy.weave(route)

    async def ack(self, route: Route) -> None:
        return await self.proxy.ack(route)

    async def disable(self, node_id: str) -> None:
        return await self.proxy.disable(node_id)

    async def enable(self, node_id: str) -> None:
        return await self.proxy.enable(node_id)

    async def index(self, index: Paging) -> Page[Route]:
        return await self.proxy.index(index)

    async def version(self, node_id: str) -> Versions:
        return await self.proxy.version(node_id)

    async def instx(self, index: Paging) -> Page[Institution]:
        return await self.proxy.instx(index)

    async def instr(self, institutions: List[Institution]) -> str:
        return await self.proxy.instr(institutions)

    async def ally(self, node_ids: List[str]) -> None:
        await self.proxy.ally(node_ids)

    async def disband(self, node_ids: List[str]) -> None:
        await self.proxy.disband(node_ids)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any


class GrpcMarshaller:

    @staticmethod
    def deserialize(buffer: Any):
        return buffer

    @staticmethod
    def serialize(buffer: Any):
        return buffer



#
# Copyright (c) 2019, 2023, ducesoft and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import random
from typing import Dict, Any

import mesh.telemetry as telemetry
import mesh.tool as tool
from mesh.context import Mesh
from mesh.macro import spi
from mesh.mpc.filter import Filter, Invoker, Invocation, CONSUMER, PROVIDER
from mesh.prsim import Metadata
from mesh.prsim import RunMode

TELEMETRY_TRACE_ID = 'mesh-telemetry-trace-id'
TELEMETRY_SPAN_ID = "mesh-telemetry-span-id"


@spi(name="telemetryConsumer", pattern=CONSUMER, priority=50 - (1 << 32))
class TelemetryConsumerFilter(Filter):

    async def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        if TELEMETRY_TRACE_ID not in invocation.get_attachments():
            invocation.get_attachments()[TELEMETRY_TRACE_ID] = hex(random.getrandbits(128))[2:]

        if not RunMode.TRACE.match(Mesh.context().get_run_mode()):
            return await invoker.run(invocation)

        attachments: Dict[str, str] = Mesh.context().get_attachments()
        if not telemetry.if_rebuild_span():
            return await invoker.run(invocation)
        with telemetry.build_via_remote(attachments, invocation.get_urn().string()):
            current_span = telemetry.get_current_span()
            current_span.set_attribute('mesh-trace-id', Mesh.context().get_trace_id())
            return await invoker.run(invocation)


@spi(name="telemetryProvider", pattern=PROVIDER, priority=(1 << 32) - 50)
class TelemetryProviderFilter(Filter):
    """
    recover telemetry context via mesh context for tracing
    """

    async def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        if TELEMETRY_TRACE_ID not in invocation.get_attachments():
            invocation.get_attachments()[TELEMETRY_TRACE_ID] = hex(random.getrandbits(128))[2:]

        if not RunMode.TRACE.match(Mesh.context().get_run_mode()):
            return await invoker.run(invocation)

        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not telemetry.if_rebuild_span():
            return await invoker.run(invocation)
        trace_id = attachments.get(Metadata.MESH_TRACE_ID.key(), tool.new_trace_id())
        with telemetry.build_via_remote(attachments, invocation.get_urn().string()):
            current_span = telemetry.get_current_span()
            current_span.set_attribute('mesh-trace-id', trace_id)
            return await invoker.run(invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import json
import os
import platform
from typing import List, Dict

import mesh.tool as tool
from mesh.kinds import Versions
from mesh.macro import spi, mps, ServiceLoader
from mesh.prsim import Builtin, Hodor


@mps
@spi("mesh")
class MeshBuiltin(Builtin):

    async def doc(self, name: str, formatter: str) -> str:
        return ""

    async def version(self) -> Versions:
        def prop() -> Dict[str, str]:
            try:
                with open(os.path.join(tool.pwd(), f"{tool.get_mesh_name()}.version")) as vfd:
                    vps = json.load(vfd)
                    return vps if vps else {}
            except FileNotFoundError:
                return {}

        properties = prop()
        name = f"{tool.get_mesh_name()}_version".upper()
        mv = tool.get_property("", [name])
        version = tool.anyone(mv, properties.get("version", "1.0.0"))
        versions = Versions()
        versions.version = version
        versions.infos = {
            f"{tool.get_mesh_name()}.commit_id": properties.get("commit_id", "3dd81bc"),
            f"{tool.get_mesh_name()}.os": platform.system(),
            f"{tool.get_mesh_name()}.arch": platform.architecture()[0],
            f"{tool.get_mesh_name()}.version": version,
        }
        return versions

    async def debug(self, features: Dict[str, str]):
        doors = ServiceLoader.load(Hodor).list('')
        if not doors:
            return
        for hodor in doors:
            await hodor.debug(features)

    async def stats(self, features: List[str]) -> Dict[str, str]:
        indies = {}
        doors = ServiceLoader.load(Hodor).list('')
        if not doors:
            return indies
        for hodor in doors:
            stats = await hodor.stats(features)
            if not stats:
                continue
            for key, value in stats.items():
                indies[key] = value
        return indies

    async def fallback(self):
        pass

    async def dump(self, names: List[str]) -> Dict[str, str]:
        return {}



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod, ABC
from typing import Any, Dict
from typing import Type, Generic

from mesh.macro import T, Inspector, Parameters, Returns
from mesh.macro import spi
from mesh.mpc.urn import URN


@spi(name='invoker', pattern="*")
class Invoker(Generic[T], ABC):
    """
    Invoke the next invoker.
    """

    @abstractmethod
    async def run(self, invocation: "Invocation") -> Any:
        pass


class Invocation(ABC):

    @abstractmethod
    def get_proxy(self) -> Invoker[T]:
        """
         Get the delegate target object.
        :return: proxy target
        """
        pass

    @abstractmethod
    def get_inspector(self) -> Inspector:
        """
        Get the invocation inspector.
        :return: inspector
        """
        pass

    @abstractmethod
    def get_parameters(self) -> Parameters:
        """
        Invoke parameters. include arguments and parameters.
        :return: parameters
        """
        pass

    @abstractmethod
    def get_arguments(self) -> [Any]:
        """
        Invoke parameters.
        :return: arguments
        """
        pass

    @abstractmethod
    def get_attachments(self) -> Dict[str, str]:
        """
        Get the attachments. The attributes will be serialized.
        :return: attachments
        """
        pass

    @abstractmethod
    def get_execution(self) -> "Execution":
        """ Get the invocation execution. """
        pass

    @abstractmethod
    def get_urn(self) -> URN:
        """
        Get the invoke urn.
        :return:
        """
        pass


class Execution(Invoker[T], Generic[T], ABC):
    """

    """

    @abstractmethod
    def schema(self) -> T:
        """ Execution schema. """
        pass

    @abstractmethod
    def inspect(self) -> Inspector:
        """ Inspect execution. """
        pass

    @abstractmethod
    def intype(self) -> Type[Parameters]:
        """ Execution input type. """
        pass

    @abstractmethod
    def retype(self) -> Type[Returns]:
        """ Execution output return type. """
        pass

    @abstractmethod
    def inflect(self) -> Parameters:
        """ Reflect input type. """
        pass

    @abstractmethod
    def reflect(self) -> Returns:
        """ Reflect output return type. """
        pass


class ServiceInvocation(Generic[T], Invocation):

    def __init__(self, proxy: Invoker[T], inspector: Inspector, parameters: Parameters, execution: Execution, urn: URN):
        self.proxy = proxy
        self.inspector = inspector
        self.parameters = parameters
        self.execution = execution
        self.urn = urn

    def get_proxy(self) -> Invoker[T]:
        return self.proxy

    def get_inspector(self) -> Inspector:
        return self.inspector

    def get_parameters(self) -> Parameters:
        return self.parameters

    def get_arguments(self) -> [Any]:
        return self.parameters.get_arguments()

    def get_attachments(self) -> Dict[str, str]:
        return self.parameters.get_attachments()

    def get_execution(self) -> Execution:
        return self.execution

    def get_urn(self) -> URN:
        return self.urn
